# SPDX-License-Identifier: MPL-2.0
# Copyright 2022 John Mille <john@ews-network.net>


from __future__ import annotations

from datetime import datetime, timedelta
from os import environ

from pynamodb.attributes import (
    ListAttribute,
    MapAttribute,
    UnicodeAttribute,
    UTCDateTimeAttribute,
)
from pynamodb.models import Model

REGISTRY_TABLE = environ.get("REGISTRY_TABLE", "certbot-registry")
REGISTRY_REGION = environ.get(
    "AWS_DEFAULT_REGION", environ.get("AWS_REGION", "eu-west-1")
)


class CertificateArns(Model):
    """
    :cvar secretsmanagerCertsArn: Mapping to each individual certificate in their own secret
    :cvar secretsmanagerArn: Secret with all the certificates in the single secret
    """

    class Meta:
        table_name = REGISTRY_TABLE
        region = REGISTRY_REGION

    hostname = UnicodeAttribute(hash_key=True)
    alt_subjects = ListAttribute(null=False)
    account_id = UnicodeAttribute(null=False)
    endpoint = UnicodeAttribute(null=False)
    expiry = UTCDateTimeAttribute(null=True)
    s3Arn = MapAttribute(
        certChain=MapAttribute(
            Url=UnicodeAttribute(null=False), Arn=UnicodeAttribute(null=False)
        ),
        fullChain=MapAttribute(
            Url=UnicodeAttribute(null=False), Arn=UnicodeAttribute(null=False)
        ),
        privateKey=MapAttribute(
            Url=UnicodeAttribute(null=False), Arn=UnicodeAttribute(null=False)
        ),
        publicKey=MapAttribute(
            Url=UnicodeAttribute(null=False), Arn=UnicodeAttribute(null=False)
        ),
        passphrase=UnicodeAttribute(null=True),
    )
    secretsmanagerCertsArn = MapAttribute(
        privateKey=UnicodeAttribute(null=False),
        publicKey=UnicodeAttribute(null=False),
        certChain=UnicodeAttribute(null=False),
        fullChain=UnicodeAttribute(null=False),
    )
    secretsmanagerArn = UnicodeAttribute(null=True)
    acmArn = UnicodeAttribute(null=True)


def get_certs_to_renew(expires_in_days: int = 28):
    """
    Returns list of certificates that have an expiry in expires_in_days.
    Defaults to 28 days, which ensures that we are past the 60 days default
    range set by Let's Encrypt

    :param int expires_in_days:
    """
    in_future: datetime = datetime.utcnow() + timedelta(days=expires_in_days)
    to_renew = CertificateArns().scan(CertificateArns.expiry < in_future)
    return to_renew



# generated by datamodel-codegen:
#   filename:  acme_config.json
#   timestamp: 2022-07-22T10:25:55+00:00

from __future__ import annotations

from typing import Any, Dict, List, Optional, Union

from pydantic import BaseModel, Extra, Field, constr


class MetaItem(BaseModel):
    uri: Optional[str] = None


class RegistrationItem(BaseModel):
    uri: Optional[str] = None


class Account(BaseModel):
    class Config:
        extra = Extra.forbid

    account_id: constr(regex=r"^\d+$")
    endpoint: str = Field(
        ..., description="ACME API endpoint the account is registered with"
    )
    created_on: Optional[str] = Field(None, description="created_dt from meta.json")
    dirname: constr(regex=r"^[a-zA-Z\d]+$") = Field(
        ..., description="The hash of the directory to store the files into"
    )
    privateKey: Optional[Union[Dict[str, Any], str]] = None
    meta: Optional[Union[MetaItem, str]] = None
    registration: Optional[Union[RegistrationItem, str]] = None

    def is_staging(self) -> bool:
        if "acme-staging" in self.endpoint:
            return True
        return False


class AcmeConfig(BaseModel):
    accounts: List[Account]



#  SPDX-License-Identifier: MPL-2.0
#  Copyright 2022 John Mille <john@ews-network.net>

"""
Lambda function to run on a schedule that will scan all the items in the DynamoDB certs registry, evaluate which ones
to renew, and trigger said renewal.
"""

import json

from boto3.session import Session
from compose_x_common.compose_x_common import set_else_none

from certbot_aws_store.registry import get_certs_to_renew


def certs_to_renew_inventory(event, context):
    """
    Main entry point for the Lambda function.

    :param event: The event passed by the AWS Lambda service.
    :param context: The context passed by the AWS Lambda service.
    :return: The response to be returned to the AWS Lambda service.
    """
    session = Session()
    expiry = int(set_else_none("CertsExpiryInDays", event, 21))
    _certs_to_renew = get_certs_to_renew(expiry)


def renew_certificate(event, context):
    """
    Function to handle a given certificate to renew by its hostname

    :param event: The event passed by the AWS Lambda service.
    :param context: The context passed by the AWS Lambda service.
    """



# SPDX-License-Identifier: MPL-2.0
# Copyright 2022 John Mille <john@ews-network.net>

"""
Swiss Knife functions to re-use across the program
"""


def easy_read(path):
    with open(path) as file:
        contents = file.read()
    return contents



#!/usr/bin/env python

"""
CLI Entrypoint for certbot-aws-store
"""
import sys
from argparse import ArgumentParser

from boto3.session import Session

from .acme_store import AcmeStore
from .certificate import AcmeCertificate


def cli_entrypoint():
    parser = ArgumentParser("Certbot store wrapper")
    parser.add_argument(
        "--secret",
        "--secret-store-arn",
        type=str,
        required=True,
        dest="secret",
        help="ACME Account Configuration secret name/ARN",
    )
    parser.add_argument(
        "--domain",
        required=True,
        action="append",
        help="Domain name for the certificate to create",
    )
    parser.add_argument(
        "--email", required=True, type=str, help="Email for the account and ToS"
    )
    parser.add_argument(
        "--register-to-acm",
        required=False,
        default=False,
        action="store_true",
        help="Creates|Updates certificate in ACM",
    )
    parser.add_argument(
        "--dry-run",
        action="store_true",
        required=False,
        default=False,
        help="By default, use ACME Staging.",
    )
    parser.add_argument(
        "--override-folder",
        type=str,
        required=False,
        default=None,
        help="Use an existing certbot folder",
    )
    parser.add_argument(
        "--profile",
        type=str,
        required=False,
        default=None,
        help="AWS Profile to use for API requests",
    )
    parser.add_argument(
        "--s3-backend-bucket-name",
        type=str,
        required=False,
        dest="bucketName",
        help="S3 bucket to store the certificate files into",
    )
    parser.add_argument(
        "--s3-backend-prefix-key",
        type=str,
        required=False,
        dest="s3_prefix_key",
        default=None,
        help="S3 Prefix path to store the certificates",
    )
    parser.add_argument(
        "--split-secrets",
        action="store_true",
        default=False,
        required=False,
        help="If set, each certificate file gets their own secret in Secrets Manager",
    )
    # parser.add_argument(
    #     "--s3-backend-set-passphrase",
    #     type=bool,
    #     required=False,
    #     action="store_true",
    #     default=False,
    #     dest="setKeyPassphrase",
    # )
    parser.add_argument(
        "--secretsmanager-backend-prefix-key",
        type=str,
        dest="secrets_prefixKey",
        default="certbot/store",
        help="SecretsManager prefix for secret name",
    )

    args = parser.parse_args()
    if args.profile:
        session = Session(profile_name=args.profile)
    else:
        session = Session()

    store_mgr = AcmeStore(args.secret, args.override_folder, session)
    certificate = AcmeCertificate(
        args.domain[0],
        store_mgr.staging_account if args.dry_run else store_mgr.account,
        subject_alts=args.domain[1:],
    )
    try:
        certificate_content = certificate.create(
            args.email, acme_store=store_mgr, staging=args.dry_run
        )
        backends: dict = {"secretsmanager": {"prefixKey": args.secrets_prefixKey}}
        if args.bucketName:
            backends["s3"]: dict = {"bucketName": args.bucketName}
            if args.s3_prefix_key:
                backends["s3"]["prefixKey"] = args.s3_prefix_key
        certificate.save_to_backends(
            backends,
            certificate_content,
            register_to_acm=args.register_to_acm,
            split_secrets=args.split_secrets,
            session=session,
        )
        print(
            certificate.hostname,
            certificate.get_account_id(store_mgr),
            certificate.serial_number,
        )
        print(certificate.urn)
        store_mgr.save()
        return 0
    except Exception as error:
        print(error)
        return 1


if __name__ == "__main__":
    sys.exit(cli_entrypoint())



# SPDX-License-Identifier: MPL-2.0
# Copyright 2022 John Mille <john@ews-network.net>

"""
Manages the config-dir/accounts settings to avoid creating new accounts for every new certificate.
"""
import datetime
import json
import os
import stat
from shutil import rmtree
from tempfile import TemporaryDirectory
from urllib.parse import urlparse

from boto3.session import Session
from compose_x_common.aws import get_session
from dateutil import parser as dateparser

from certbot_aws_store.acme_config import Account, AcmeConfig
from certbot_aws_store.backends import SECRET_ARN_RE
from certbot_aws_store.utils import easy_read


class AcmeStore:
    pkey_file_name = "private_key.json"
    regr_file_name = "regr.json"
    meta_file_name = "meta.json"

    files = {
        pkey_file_name: "privateKey",
        regr_file_name: "registration",
        meta_file_name: "metadata",
    }

    accounts_dir_name: str = "accounts"

    def __init__(
        self,
        certs_store_arn: str,
        override_directory: str = None,
        session: Session = None,
    ):
        self._store_arn = certs_store_arn
        self.session = get_session(session)
        self.staging_account = None
        self.account = None
        self.backup_config = None
        if not override_directory:
            self.temp_dir = TemporaryDirectory()
            self.directory = self.temp_dir.name
            self.accounts_path = f"{self.config_dir}/{self.accounts_dir_name}"
            os.makedirs(self.accounts_path, exist_ok=True)
        else:
            self.directory = override_directory
            try:
                self.accounts_path = find_accounts_dir(override_directory)
            except OSError:
                self.accounts_path = f"{self.config_dir}/{self.accounts_dir_name}"
                if not os.path.exists(self.accounts_path):
                    os.makedirs(self.accounts_path, exist_ok=True)
        self.init_pull()

    @property
    def existing_accounts(self) -> list[Account]:
        return get_acme_accounts(self.accounts_path)

    @property
    def config_dir(self) -> str:
        return f"{self.directory}/config-dir"

    @property
    def logs_dir(self) -> str:
        return f"{self.directory}/logs-dir"

    @property
    def work_dir(self) -> str:
        return f"{self.directory}/work-dir"

    @property
    def config(self) -> AcmeConfig:
        accounts = get_acme_accounts(self.accounts_path)
        return AcmeConfig(accounts=accounts)

    def init_pull(
        self,
    ):
        client = self.session.client("secretsmanager")
        try:
            config_r = client.get_secret_value(SecretId=self.secret_name)
            config_content = json.loads(config_r["SecretString"])
            config_content["store_arn"] = config_r["ARN"]
            self._store_arn = config_r["ARN"]
            config = AcmeConfig(**config_content)
            self.layout_accounts_folders(config)
        except client.exceptions.ResourceNotFoundException:
            print(
                f"Secret {self._store_arn} does not exist. Will initialize after first certificate"
            )
        except Exception as error:
            print(error)
            raise
        self.set_execution_accounts()

    def layout_accounts_folders(self, config: AcmeConfig):
        for endpoint, accounts in accounts_per_endpoints(config).items():
            directory_path = create_account_endpoint_dirs(self.accounts_path, endpoint)
            for account in accounts:
                _account_path = f"{directory_path}/{account.dirname}"
                os.makedirs(_account_path, exist_ok=True)
                write_accounts_files(_account_path, account)
        self.backup_config = AcmeConfig(accounts=get_acme_accounts(self.accounts_path))

    def set_execution_accounts(self):
        now = datetime.datetime.now(tz=None)
        for endpoint, accounts in accounts_per_endpoints(self.config).items():
            latest_account = accounts[0]
            latest_account_path = (
                f"{self.config_dir}/{self.accounts_dir_name}"
                f"/{endpoint}/directory/{accounts[0].dirname}"
            )
            diff = now - get_meta_creation_dt(latest_account)
            to_clear: list = []
            for account in accounts:
                _account_path = (
                    f"{self.config_dir}/{self.accounts_dir_name}"
                    f"/{endpoint}/directory/{account.dirname}"
                )
                to_clear.append(_account_path)
                created_on = get_meta_creation_dt(account)
                if not diff or (diff and diff > (now - created_on)):
                    diff = now - created_on
                    latest_account = account
                    latest_account_path = _account_path
            set_latest_endpoint_account(
                endpoint, latest_account, latest_account_path, to_clear
            )
            if latest_account.is_staging():
                self.staging_account = latest_account
            else:
                self.account = latest_account

    @property
    def secret_name(self) -> str:
        if SECRET_ARN_RE.match(self._store_arn):
            return SECRET_ARN_RE.match(self._store_arn).group("name")
        else:
            return self._store_arn

    def save(self):
        client = self.session.client("secretsmanager")
        secret_value = AcmeConfig(
            accounts=self.merge_used_accounts_with_backup()
        ).json()
        print(
            "Saving accounts",
            [
                (_act.account_id, _act.endpoint)
                for _act in self.merge_used_accounts_with_backup()
            ],
        )
        try:
            client.put_secret_value(SecretId=self._store_arn, SecretString=secret_value)
        except client.exceptions.ResourceNotFoundException:
            client.create_secret(
                Name=self.secret_name,
                SecretString=secret_value,
                Tags=[
                    {"Key": "Name", "Value": self.secret_name},
                    {"Key": "certbot_route53_store", "Value": str(True)},
                ],
            )
        except Exception as error:
            print(error)

    def merge_used_accounts_with_backup(self) -> list[Account]:
        if not self.backup_config or not isinstance(self.backup_config, AcmeConfig):
            return self.config.accounts
        at_del_accounts = self.config.accounts
        to_add: list = []
        for account in at_del_accounts:
            for _bkp_account in self.backup_config.accounts:
                if account.endpoint != _bkp_account.endpoint:
                    continue
                if account.account_id == _bkp_account.account_id:
                    continue
                to_add.append(_bkp_account)
        at_del_accounts += to_add
        return at_del_accounts


def create_account_endpoint_dirs(accounts_path: str, endpoint: str) -> str:
    endpoint_path = f"{accounts_path}/{endpoint}"
    if not os.path.exists(endpoint_path):
        os.makedirs(endpoint_path, exist_ok=True)
    os.chmod(f"{endpoint_path}", stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR)
    directory_path = f"{endpoint_path}/directory"
    if not os.path.exists(directory_path):
        os.makedirs(directory_path, exist_ok=True)
    os.chmod(f"{directory_path}", stat.S_IWUSR | stat.S_IRUSR | stat.S_IXUSR)
    return directory_path


def set_latest_endpoint_account(
    endpoint: str, account: Account, account_path: str, to_clear: list
):
    print(f"Latest {endpoint} account: {account.dirname} - {account.account_id}")
    for _dir_to_remove in to_clear:
        if _dir_to_remove == account_path:
            continue
        rmtree(_dir_to_remove)
    if not os.path.exists(account_path):
        os.makedirs(account_path, exist_ok=True)
    elif os.path.exists(account_path) and not os.path.isdir(account_path):
        raise OSError(account_path, "exists but is not a directory")


def write_accounts_files(account_path: str, account: Account) -> None:
    """
    Creates the account folders, writes down the private_key.json and regr.json
    """
    print("Import for account", account.account_id, account.endpoint, account_path)
    private_key_path = f"{account_path}/{AcmeStore.pkey_file_name}"
    try:
        with open(private_key_path) as fd:
            key_content = fd.read()
            if set(key_content) == set(account.privateKey):
                print(
                    private_key_path, "already exists and is identical. Nothing to do"
                )
    except OSError:
        with open(private_key_path, "w") as fd:
            if isinstance(account.privateKey, str):
                fd.write(account.privateKey)
            elif isinstance(account.privateKey, dict):
                fd.write(json.dumps(account.privateKey))
        os.chmod(private_key_path, stat.S_IRUSR)

    regr_key_path = f"{account_path}/{AcmeStore.regr_file_name}"
    try:
        with open(regr_key_path) as fd:
            key_content = fd.read()
            if set(key_content) == set(account.registration):
                print(regr_key_path, "already exists and is identical. Nothing to do")
    except OSError:
        with open(regr_key_path, "w") as fd:
            if isinstance(account.registration, str):
                fd.write(account.registration)
            elif isinstance(account.registration, dict):
                fd.write(json.dumps(account.registration))
        os.chmod(
            regr_key_path,
            stat.S_IWRITE | stat.S_IRUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH,
        )

    meta_key_path = f"{account_path}/{AcmeStore.meta_file_name}"
    try:
        with open(meta_key_path) as fd:
            key_content = fd.read()
            if set(key_content) == set(account.meta):
                print(meta_key_path, "already exists and is identical. Nothing to do")
    except OSError:
        with open(meta_key_path, "w") as fd:
            if isinstance(account.meta, str):
                fd.write(account.meta)
            elif isinstance(account.meta, dict):
                fd.write(json.dumps(account.meta))
        os.chmod(
            meta_key_path,
            stat.S_IWRITE | stat.S_IRUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH,
        )


def find_accounts_dir(certs_root_path, accounts_folder_name: str = "accounts") -> str:
    """
    For a given path, returns the path to ``accounts`` folder.

    :raises: IOError if ``accounts_folder_name`` not found in tree
    """
    for root, dirs, files in os.walk(certs_root_path):
        for _dir in dirs:
            if _dir == accounts_folder_name:
                dir_path = os.path.join(root, _dir)
                if not os.path.isdir(dir_path):
                    raise TypeError(dir_path, "is not a directory, yet exists.")
                return dir_path
    raise OSError(
        f"No folder named {accounts_folder_name} found in {certs_root_path} tree"
    )


def get_acme_accounts(
    accounts_path: str,
) -> list[Account]:
    """

    :param str accounts_path:
    :return: list of accounts
    :rtype: list[Account]
    """
    accounts: list = []
    for root, dirs, files in os.walk(accounts_path):
        if AcmeStore.pkey_file_name not in files:
            continue
        for file in files:
            if file == AcmeStore.pkey_file_name:
                private_key_path = os.path.join(root, file)
                top_dir = os.path.split(private_key_path)[0]
                top_path = os.path.normpath(top_dir)
                reg_key_path = os.path.join(top_dir, AcmeStore.regr_file_name)
                meta_key_path = os.path.join(top_dir, AcmeStore.meta_file_name)
                account_parts = top_path.split(os.sep)[3:]
                reg_key_content = easy_read(reg_key_path)
                meta_content = easy_read(meta_key_path)
                account_uri = json.loads(reg_key_content)["uri"]
                acme_account = urlparse(account_uri)
                acme_account_id = os.path.split(acme_account.path)[-1]
                account = Account(
                    account_id=acme_account_id,
                    privateKey=easy_read(private_key_path),
                    registration=reg_key_content,
                    meta=meta_content,
                    created_on=datetime.datetime.replace(
                        dateparser.parse(json.loads(meta_content)["creation_dt"]),
                        tzinfo=None,
                    ).isoformat(),
                    dirname=account_parts[-1],
                    endpoint=acme_account.netloc,
                )
                accounts.append(account)
    return accounts


def get_meta_creation_dt(account: Account) -> datetime.datetime:
    if isinstance(account.meta, str):
        metadata = json.loads(account.meta)
        created_on = datetime.datetime.replace(
            dateparser.parse(metadata["creation_dt"]), tzinfo=None
        )
    elif isinstance(account.meta, dict):
        created_on = datetime.datetime.replace(
            dateparser.parse(account.meta["creation_dt"]), tzinfo=None
        )
    else:
        raise TypeError(
            "account.meta is not valid type. Got",
            type(account.meta),
            "expected one of",
            (str, dict),
        )
    return created_on


def accounts_per_endpoints(config: AcmeConfig) -> dict[str, list[Account]]:
    endpoints_to_accounts: dict = {}
    for account in config.accounts:
        if account.endpoint not in endpoints_to_accounts.keys():
            endpoints_to_accounts[account.endpoint]: list = [account]
        else:
            endpoints_to_accounts[account.endpoint].append(account)
    return endpoints_to_accounts



# SPDX-License-Identifier: MPL-2.0
# Copyright 2022 John Mille <john@ews-network.net>

"""
Manages the provisioning and storage of the ACME/Let's Encrypt certificate
"""

from __future__ import annotations

import datetime
import json
import os.path
import re
import warnings
from typing import TYPE_CHECKING, Union

if TYPE_CHECKING:
    from certbot_aws_store.acme_store import AcmeStore
    from certbot_aws_store.acme_config import Account

from multiprocessing import Process

import OpenSSL
from b_dynamodb_common.models.model_type_factory import ModelTypeFactory
from boto3.session import Session
from certbot.main import main
from compose_x_common.aws import get_session
from compose_x_common.aws.acm import find_certificate_from_domain_name
from compose_x_common.compose_x_common import set_else_none
from dateutil import parser

from certbot_aws_store.backends import (
    handle_s3_backend,
    handle_secretsmanager_secret_all_certs,
    handle_secretsmanager_secret_per_cert,
    pull_from_s3,
    pull_from_secrets_manager_aio,
    pull_from_secretsmanager_secret_per_cert,
)
from certbot_aws_store.registry import REGISTRY_REGION, REGISTRY_TABLE, CertificateArns
from certbot_aws_store.utils import easy_read


def find_certificate_renewal_config(top_path: str, hostname: str) -> list:
    renewal_config_file_name = f"{hostname}.conf"
    for root, dirs, files in os.walk(top_path):
        for _file in files:
            if _file == renewal_config_file_name:
                file_path = os.path.join(root, _file)
                with open(file_path) as config_fd:
                    return config_fd.readlines()


def provision_cert(
    email,
    domains: list[str],
    dir_path: str,
    staging: bool = False,
    accept_tos: bool = True,
    config_dir: str = None,
    logs_dir: str = None,
    work_dir: str = None,
):
    """
    Provisions a new certificate with Route53 validation
    """
    cmd = [
        "certonly",
        "-n",
        "--dns-route53",
        "-d",
        ",".join(domains),
        "--config-dir",
        config_dir or f"{dir_path}/config-dir/",
        "--work-dir",
        work_dir or f"{dir_path}/work-dir/",
        "--logs-dir",
        logs_dir or f"{dir_path}/logs-dir/",
    ]
    if staging:
        cmd.append("--test-cert")
    if accept_tos:
        cmd.append("--agree-tos")
        cmd.append("--email")
        cmd.append(email)
    main(cmd)


class AcmeCertificate:
    certificate_file_name: str = "cert.pem"
    private_key_file_name: str = "privkey.pem"
    chain_file_name: str = "chain.pem"
    full_chain_file_name: str = "fullchain.pem"

    files: dict = {
        certificate_file_name: "publicKey",
        chain_file_name: "certChain",
        private_key_file_name: "privateKey",
        full_chain_file_name: "fullChain",
    }

    default_prefix: str = "certbot/store/"

    def __init__(
        self,
        hostname,
        acme_store_account: Account,
        subject_alts: list = None,
        table_name: str = None,
        region_name: str = None,
    ):
        """
        Init the certificate. If all you need is download, you can set acme_store_account to None

        :param str hostname:
        :param AcmeStore acme_store_account:
        :param list[str] subject_alts:
        :param str table_name:
        :param str region_name:
        """
        self._hostname = hostname
        self.alt_subjects = subject_alts if subject_alts else []
        self.registry_table = ModelTypeFactory(CertificateArns).create(
            custom_table_name=table_name or REGISTRY_TABLE,
            custom_region=region_name or REGISTRY_REGION,
        )
        self.registry_cert = self.registry_table(hash_key=self.hostname)
        self.s3_backend = None
        self.secretsmanager_backend = None
        self.certs_paths: dict = {}
        self.renewal_config: list = []
        self.acme_account = acme_store_account

    def __repr__(self):
        return self.urn

    @property
    def hostname(self) -> str:
        return self._hostname

    @property
    def hostnames(self) -> list[str]:
        """
        List with the main domain name and subject alternative names
        """
        return [self.hostname] + self.alt_subjects

    @property
    def urn(self) -> str:
        if self.exists():
            props = json.loads(self.get())
            return f"{props['endpoint']}::{props['account_id']}::{self.hostname}"
        return f"{self.acme_account.endpoint}::{self.acme_account.account_id}::{self.hostname}"

    def exists(self):
        if not self.registry_cert.exists():
            self.registry_cert.create_table()
            return False
        try:
            self.registry_cert.get(hash_key=self.hostname)
            return True
        except self.registry_cert.DoesNotExist:
            return False

    def get(self):
        cert = self.registry_cert.get(hash_key=self.hostname)
        return cert.to_json()

    def create(self, email: str, acme_store: AcmeStore, staging: bool = False):
        if self.acme_account:
            args = (email, self.hostnames, acme_store.directory, staging)
        else:
            args = (email, self.hostnames, acme_store.directory, staging, True)
        process = Process(
            target=provision_cert,
            args=args,
        )
        process.start()
        process.join()
        live_path = f"{acme_store.config_dir}/live/{self.hostname}"
        certificate_files: dict = {}
        for file, attribute in self.files.items():
            file_path = f"{live_path}/{file}"
            certificate_files[file]: str = easy_read(file_path)
            self.certs_paths[attribute]: str = os.path.abspath(file_path)
        self.renewal_config = find_certificate_renewal_config(
            acme_store.directory, self.hostname
        )
        if not self.acme_account:
            acme_store.set_execution_accounts()
            self.acme_account = (
                acme_store.staging_account if staging else acme_store.account
            )
        return certificate_files

    @property
    def certbot_account_id(self) -> Union[str, None]:
        for _line in self.renewal_config:
            if _line.startswith("account"):
                certbot_account_id = re.sub(
                    r"[^a-zA-Z\d]+", r"", _line.split(r"=")[-1].strip()
                )
                break
        else:
            print("Unable to get the renewal information")
            return None
        return certbot_account_id

    def get_account_id(self, acme_store: AcmeStore):
        certbot_account_id = self.certbot_account_id
        if not certbot_account_id:
            return None
        for account in acme_store.config.accounts:
            if account.dirname == certbot_account_id:
                return account.account_id

    @property
    def public_key_x509(self):
        if self.files[self.certificate_file_name] not in self.certs_paths:
            return
        return OpenSSL.crypto.load_certificate(
            OpenSSL.crypto.FILETYPE_PEM,
            easy_read(self.certs_paths[self.files[self.certificate_file_name]]),
        )

    @property
    def expiry_date(self) -> Union[None, datetime.datetime]:
        if not self.public_key_x509:
            warnings.warn(
                UserWarning("Unable to determine expiry date from public key.")
            )
            return None
        return parser.parse(self.public_key_x509.get_notAfter())

    @property
    def serial_number(self) -> Union[str, None]:
        if not self.public_key_x509:
            warnings.warn(
                UserWarning("Unable to determine expiry date from public key.")
            )
            return None
        return self.public_key_x509.get_serial_number()

    def save_to_s3(
        self,
        s3_registry_arn: dict,
        backend_definition: dict,
        cert_content: dict,
        session: Session = None,
    ):
        for file_name, attribute_name in self.files.items():
            s3_backend = handle_s3_backend(
                self,
                file_name,
                cert_content,
                backend_definition["bucketName"],
                set_else_none(
                    "prefixKey", backend_definition, alt_value=self.default_prefix
                ),
                session,
            )
            s3_registry_arn.update(
                {attribute_name: {"Arn": s3_backend.arn, "Url": s3_backend.url}}
            )

    def save_to_secretsmanager(
        self,
        split_secrets: bool,
        cert_content: dict,
        backend_definition: dict,
        session: Session = None,
    ) -> Union[str, dict]:
        session = get_session(session)
        if split_secrets:
            secretsmanager_registry_arn: dict = {}
            for file_name, attribute_name in self.files.items():
                secretsmanager_backends = handle_secretsmanager_secret_per_cert(
                    self,
                    file_name,
                    cert_content,
                    prefix_key=set_else_none(
                        "prefixKey",
                        backend_definition,
                        alt_value=self.default_prefix,
                    ),
                    session=session,
                )
                secretsmanager_registry_arn.update(
                    {
                        attribute_name: secretsmanager_backends.arn,
                    }
                )
            return secretsmanager_registry_arn
        else:
            secretsmanager_backends = handle_secretsmanager_secret_all_certs(
                self,
                cert_content,
                prefix_key=set_else_none(
                    "prefixKey", backend_definition, alt_value=self.default_prefix
                ),
                session=session,
            )
            return secretsmanager_backends.arn

    def save_to_backends(
        self,
        backends: dict,
        cert_content: dict,
        register_to_acm: bool = True,
        split_secrets: bool = False,
        session: Session = None,
    ):
        session = get_session(session)
        s3_registry_arn: dict = {}
        secretsmanager_registry_arn: dict = {}
        for backend_name, backend_definition in backends.items():
            if backend_name == "s3":
                try:
                    self.save_to_s3(
                        s3_registry_arn, backend_definition, cert_content, session
                    )
                except Exception as error:
                    print(error)
                    print("Failed to save certificate to S3")
            elif backend_name == "secretsmanager":
                try:
                    secretsmanager_registry_arn = self.save_to_secretsmanager(
                        split_secrets,
                        cert_content,
                        backend_definition,
                        session,
                    )
                except Exception as error:
                    print(error)
                    print("Failed to save certificate to Secrets manager")

        acm_arn = None
        if register_to_acm:
            try:
                acm_arn = self.register_to_acm(cert_content, session)
            except Exception as error:
                print(error)
                print("Failed to save certificate to ACM")

        self.register_to_registry(s3_registry_arn, secretsmanager_registry_arn, acm_arn)

    def register_to_registry(
        self,
        s3_registry_arn: dict = None,
        secretsmanager_registry_arn: Union[str, dict] = None,
        acm_arn: str = None,
    ):
        props: dict = {
            "hostname": self.hostname,
            "expiry": self.expiry_date,
            "account_id": self.acme_account.account_id,
            "endpoint": self.acme_account.endpoint,
            "alt_subjects": self.alt_subjects,
        }
        if s3_registry_arn:
            props["s3Arn"] = s3_registry_arn
        if secretsmanager_registry_arn and isinstance(
            secretsmanager_registry_arn, dict
        ):
            props["secretsmanagerCertsArn"] = secretsmanager_registry_arn
        else:
            props["secretsmanagerArn"] = secretsmanager_registry_arn
            props["secretsmanagerCertsArn"]: dict = {}
        if acm_arn:
            props["acmArn"] = acm_arn
        registry_entry = CertificateArns(**props)

        registry_entry.save()
        return registry_entry.to_json()

    def register_to_acm(self, cert_content: dict, session: Session = None):
        session = get_session(session)
        certificate = find_certificate_from_domain_name(self.hostname, True, session)
        client = session.client("acm")
        if certificate:
            arn = certificate["CertificateArn"]
            client.import_certificate(
                CertificateArn=arn,
                Certificate=cert_content[self.certificate_file_name],
                PrivateKey=cert_content[self.private_key_file_name],
                CertificateChain=cert_content[self.chain_file_name],
            )
        else:
            cert_r = client.import_certificate(
                Certificate=cert_content[self.certificate_file_name],
                PrivateKey=cert_content[self.private_key_file_name],
                CertificateChain=cert_content[self.chain_file_name],
                Tags=[{"Key": "Name", "Value": self.hostname}],
            )
            arn = cert_r["CertificateArn"]
        return arn

    def pull(
        self,
        destination_folder: str,
        use_s3: bool = False,
        split_secrets: bool = False,
        session: Session = None,
    ):
        session = get_session(session)
        if not self.exists():
            raise ValueError(self.hostname, f"not found in table")
        details = json.loads(self.registry_cert.get(self.hostname).to_json())
        if use_s3:
            locations = set_else_none("s3Arn", details)
        elif split_secrets:
            locations = set_else_none("secretsmanagerCertsArn", details)
        else:
            locations = set_else_none("secretsmanagerArn", details)

        if not locations:
            raise ValueError(
                "Unable to determine the location of certificates from dynamodb registry"
            )

        if isinstance(locations, str) and not use_s3:
            pull_from_secrets_manager_aio(
                destination_folder, locations, self.private_key_file_name, session
            )
        elif isinstance(locations, dict) and use_s3:
            pull_from_s3(
                destination_folder, locations, self.private_key_file_name, session
            )
        else:
            pull_from_secretsmanager_secret_per_cert(
                destination_folder, locations, self.private_key_file_name, session
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['certbot_aws_store']

package_data = \
{'': ['*']}

install_requires = \
['argparse>=1.4.0,<2.0.0',
 'b-dynamodb-common>=0.4.0,<0.5.0',
 'certbot-dns-route53>=2.3,<3.0',
 'certbot>=2.3.0,<3.0.0',
 'compose-x-common>=1.2,<2.0',
 'pyOpenSSL>=22.0.0,<23.0.0',
 'pynamodb>=5.2.1,<6.0.0']

entry_points = \
{'console_scripts': ['certbot-aws-store = '
                     'certbot_aws_store.cli:cli_entrypoint']}

setup_kwargs = {
    'name': 'certbot-aws-store',
    'version': '0.4.0',
    'description': "Generate Let's Encrypt certificates and store into AWS",
    'long_description': '==================================\ncertbot-aws-store\n==================================\n\nWrapper tool / function around certbot + route53 that allows to request new certificates, and store them in AWS.\n\nFeatures\n\n* Store Let\'s Encrypt certificates in AWS SecretsManager (default) and S3 (optional)\n* Keep track of ACME account configuration and store securely in AWS SecretsManager\n* Keep track of certificates issued and stored in account via DynamoDB\n\nPre-Requisites\n================\n\n* A function AWS Account and credentials to make API calls\n* Access to DynamoDB + SecretsManager + S3 (optional)\n\nWe recommend to create the dynamodb table using the ``certbot-store-registry.template`` CFN template, otherwise\ncertbot-aws-store will attempt to create it programmatically.\n\nThe needed RCU/WCU should be 5/5 (default) or lower.\n\nInstall\n=========\n\n.. code-block::\n\n    pip install certbot-aws-store --user\n\n    python3 -m venv venv\n    source venv/bin/activate\n    pip install pip -U; pip install certbot-aws-store\n\nUsage\n======\n\nAs a CLI\n----------\n\n\n.. code-block::\n\n    usage: Certbot store wrapper [-h] --secret SECRET --domain DOMAIN --email EMAIL [--register-to-acm] [--dry-run] [--override-folder OVERRIDE_FOLDER] [--profile PROFILE] [--s3-backend-bucket-name BUCKETNAME]\n                                 [--s3-backend-prefix-key S3_PREFIX_KEY] [--split-secrets] [--secretsmanager-backend-prefix-key SECRETS_PREFIXKEY]\n\n    optional arguments:\n      -h, --help            show this help message and exit\n      --secret SECRET, --secret-store-arn SECRET\n                            ACME Configuration secret name/ARN\n      --domain DOMAIN       Domain name for the certificate to create\n      --email EMAIL         Email for the account and ToS\n      --register-to-acm     Creates|Updates certificate in ACM\n      --dry-run             By default, use ACME Staging.\n      --override-folder OVERRIDE_FOLDER\n                            Use an existing certbot folder\n      --profile PROFILE     AWS Profile to use for API requests\n      --s3-backend-bucket-name BUCKETNAME\n                            S3 bucket to store the certificate files into\n      --s3-backend-prefix-key S3_PREFIX_KEY\n                            S3 Prefix path to store the certificates\n      --split-secrets       If set, each certificate file gets their own secret in Secrets Manager\n      --secretsmanager-backend-prefix-key SECRETS_PREFIXKEY\n                            SecretsManager prefix for secret name\n\nExample\n--------\n\n.. code-block::\n\n    certbot-aws-store --secret dev-acme-store --override-folder certbot-store \\\n     --email john@ews-network.net \\\n    --domain test-local-0005.bdd-testing.compose-x.io \\\n    --s3-backend-bucket-name dev-test-bucket\n\n\nInspiration\n=============\n\nLet\'s Encrypt + Certbot is a goto for anyone who wishes to have free SSL certificates to use in various places.\nBut then the certificates management, storage, backup and so on, still has to be done.\n\nThis is an attempt at automating the storage of certificates in AWS and the associated ACME account configuration\n(to avoid rate limiting).\n\nThis tool can be used as a CLI, and coming soon, an AWS Lambda Function or/and (coming soon) a CloudFormation resource.\nOnce installed on AWS, the registry will be automatically looked at daily to identify certificates that need to be\nrenewed and store the new values in appropriate places.\n\nHow does it work ?\n=====================\n\nOn the first time, if the ACME secret does not exist, we consider you never used ``certbot-aws-store`` before,\nand a new ACME account will be created, along with the certificate requested.\n\nOnce the certificate request is successfully completed, both the certificate and the ACME account details are saved\nto secrets manager (the certificate)\n\nUsing the dynamoDB "registry" table, we store the ARN to the various files stored in AWS, along with some metadata.\n\nFor example, the following represents a certificate stored in Secrets Manager, S3 and ACM\n\n.. code-block:: json\n\n    {\n     "hostname": "test-local-0106.bdd-testing.compose-x.io",\n     "account_id": "61811954",\n     "acmArn": "arn:aws:acm:eu-west-1:373709667837:certificate/82b3ab6f-5b53-4a3b-ab7d-ccd1ecb52255",\n     "endpoint": "acme-staging-v02.api.letsencrypt.org",\n     "expiry": "2022-10-20T22:38:59.000000+0000",\n     "s3Arn": {\n      "certChain": "arn:aws:s3:::dev-test-bucket::certbot/store/acme-staging-v02.api.letsencrypt.org/61812954/test-local-0106.bdd-testing.compose-x.io/chain.pem",\n      "fullChain": "arn:aws:s3:::dev-test-bucket::certbot/store/acme-staging-v02.api.letsencrypt.org/61812954/test-local-0106.bdd-testing.compose-x.io/fullchain.pem",\n      "privateKey": "arn:aws:s3:::dev-test-bucket::certbot/store/acme-staging-v02.api.letsencrypt.org/61812954/test-local-0106.bdd-testing.compose-x.io/privkey.pem",\n      "publicKey": "arn:aws:s3:::dev-test-bucket::certbot/store/acme-staging-v02.api.letsencrypt.org/61812954/test-local-0106.bdd-testing.compose-x.io/cert.pem"\n     },\n     "secretsmanagerArn": "arn:aws:secretsmanager:eu-west-1:373709667837:secret:certbot/store/acme-staging-v02.api.letsencrypt.org/61812954/test-local-0106.bdd-testing.compose-x.io-HpgyTD",\n     "secretsmanagerCertsArn": {}\n    }\n\nThe registry will be used in the future to evaluate / list the certificates that we have and decide whether or not\na certificate should be renewed.\n\nWhen stored in SecretsManager, we might implement a Lambda function to implement the rotation which would update\neverything, including S3.\n\n.. warning::\n\n    If you use ``--dry-run`` to use the ACME staging endpoint for testing, and request the same domain name as for\n    the production ACME endpoint, and store the certificate to ACM, the latest of the two updates the ACM certificate.\n\n\nWhy "bother" ?\n===============\n\nWith certbot, per account you get 50 certificates requests per week. Which might sound low, but then is even lower when\nyou consider the constraints of other limits.\n\nSo of course, considering a world of microservices where you might have 100s of containers needing certificates at the\nsame time, you would breach that limit in no time. So you store them centrally somewhere.\n\nRetrieving the same certificates consistently also will address issues you might have for your clients if you enable\nfeatures such as HSTS (if you do, allow for rotation within the expiry of the certificates!).\n',
    'author': 'John Preston',
    'author_email': 'john@ews-network.net',
    'maintainer': 'John Preston',
    'maintainer_email': 'john@ews-network.net',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



# SPDX-License-Identifier: MPL-2.0
# Copyright 2022 John Mille <john@ews-network.net>

"""
AWS Storage backend wrappers to make it easier to manipulate values.
"""

from __future__ import annotations

import datetime
import json
import stat
from os import chmod
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from certbot_aws_store.certificate import AcmeCertificate

import re
import warnings

from boto3.session import Session
from compose_x_common.aws import get_session
from compose_x_common.aws.s3 import S3_BUCKET_ARN_RE
from compose_x_common.compose_x_common import set_else_none

SECRETS_MANAGER_RE = re.compile(
    r"arn:aws(?:[\w-]+)?:secretsmanager:(?P<region>[a-z\d\-]+-\d):"
)
SECRET_ARN_RE = re.compile(
    rf"{SECRETS_MANAGER_RE.pattern}"
    r"(?P<accountid>[0-9]{12}):secret:(?P<id>(?P<name>[\S]+)(?:-[A-Za-z\d]{6}))$"
)


class AWSBackend:
    """
    Generic class for storage of the certificates
    """

    def __init__(self, prefix_path: str = "certbot-store/", session: Session = None):
        self._arn = ""
        self._prefix = prefix_path
        self.session = get_session(session)

    @property
    def arn(self) -> str:
        return self._arn

    @property
    def prefix(self) -> str:
        if self._prefix.endswith(r"/"):
            return self._prefix
        return f"{self._prefix}/"

    @arn.setter
    def arn(self, value: str):
        self._arn = self.set_arn(value)

    def set_arn(self, arn: str):
        return arn


class SecretsManagerBackend(AWSBackend):
    """
    SecretsManager storage backend
    """

    @property
    def arn(self) -> str:
        return self._arn

    @arn.setter
    def arn(self, value: str):
        self._arn = self.set_arn(value)

    @property
    def secret_name(self) -> str:
        if SECRETS_MANAGER_RE.match(self.arn) and SECRET_ARN_RE.match(self.arn):
            return SECRET_ARN_RE.match(self.arn).group("name")
        return self.arn

    def put(self, value, restore: bool = True):
        client = self.session.client("secretsmanager")
        try:
            existing_r = client.describe_secret(SecretId=self.secret_name)
            self.arn = existing_r["ARN"]
            planned_deletion = set_else_none("DeletedDate", existing_r)
            if planned_deletion and planned_deletion > datetime.datetime.utcnow():
                if restore:
                    warnings.warn(
                        UserWarning(
                            "Secret is planned for deletion. Cancelling deletion"
                        )
                    )
                    client.restore_secret(SecretId=self.arn)
                else:
                    warnings.warn(
                        UserWarning(
                            "Secret is planned for deletion. Not updating value!!!"
                        )
                    )
            client.put_secret_value(SecretId=self.secret_name, SecretString=value)
        except client.exceptions.ResourceNotFoundException:
            self.arn = client.create_secret(Name=self.secret_name, SecretString=value)[
                "ARN"
            ]


class S3Backend(AWSBackend):
    """
    Class to handle S3 backend storage
    """

    bucket_file_arn_re = re.compile(
        r"^arn:aws(?:-[a-z]+)?:s3:::(?P<bucket>\S+)::(?P<key>\S+)$"
    )

    def __init__(self, prefix_path: str = "certbot-store/", session: Session = None):
        self.bucket_name = ""
        self.key = ""
        super().__init__(prefix_path, session)

    def set_arn(self, arn: str):
        parts = self.bucket_file_arn_re.match(arn)
        if not parts:
            raise ValueError(
                "S3 ARN is not valid. Got",
                arn,
                "Expected to match",
                S3_BUCKET_ARN_RE.pattern,
            )
        self.bucket_name = parts.group("bucket")
        self.key = parts.group("key")
        return arn

    @property
    def url(self) -> str:
        return f"s3://{self.bucket.name}/{self.key}"

    @property
    def bucket(self):
        if not self.bucket_name:
            warnings.warn("You must set the bucket name first")
            return
        resource = self.session.resource("s3")
        return resource.Bucket(self.bucket_name)

    @property
    def object(self):
        if not self.bucket or not self.key:
            warnings.warn("You must set the bucket and key first")
            return
        resource = self.session.resource("s3")
        return resource.Object(self.bucket.name, self.key)


def handle_s3_backend(
    certificate: AcmeCertificate,
    key_name: str,
    cert_content: dict,
    s3_bucket_name,
    s3_prefix_key: str,
    session: Session = None,
) -> S3Backend:
    """
    Creates the S3 backend store for public and private key. Uploads to S3.
    """
    session = get_session(session=session)
    key_backend = S3Backend(prefix_path=s3_prefix_key, session=session)
    key_backend.arn = (
        f"arn:aws:s3:::{s3_bucket_name}"
        f"::{key_backend.prefix}{certificate.acme_account.endpoint}/"
        f"{certificate.acme_account.account_id}/{certificate.hostname}/{key_name}"
    )
    key_backend.object.put(Body=cert_content[key_name])
    return key_backend


def handle_secretsmanager_secret_per_cert(
    cert: AcmeCertificate,
    key_name: str,
    cert_content: dict,
    prefix_key: str,
    session: Session = None,
) -> SecretsManagerBackend:
    session = get_session(session)
    key_backend = SecretsManagerBackend(prefix_path=prefix_key, session=session)
    key_backend.arn = (
        f"{key_backend.prefix}"
        f"{cert.acme_account.endpoint}/{cert.acme_account.account_id}/"
        f"{cert.hostname}/{key_name}"
    )
    key_backend.put(cert_content[key_name])
    return key_backend


def handle_secretsmanager_secret_all_certs(
    cert: AcmeCertificate,
    cert_content: dict,
    prefix_key: str,
    session: Session = None,
) -> SecretsManagerBackend:
    content: dict = {}
    for file, attribute in cert.files.items():
        content[attribute] = cert_content[file]
    session = get_session(session)
    key_backend = SecretsManagerBackend(prefix_path=prefix_key, session=session)
    key_backend.arn = (
        f"{key_backend.prefix}"
        f"{cert.acme_account.endpoint}/{cert.acme_account.account_id}/"
        f"{cert.hostname}"
    )
    key_backend.put(json.dumps(content))
    return key_backend


def pull_from_secrets_manager_aio(
    destination_folder: str,
    secret_arn: str,
    private_file_name: str = None,
    session: Session = None,
) -> None:
    from certbot_aws_store.certificate import AcmeCertificate

    session = get_session(session)
    client = session.client("secretsmanager")
    try:
        secret_value = json.loads(
            client.get_secret_value(SecretId=secret_arn)["SecretString"]
        )
    except client.exceptions.ResourceNotFoundException as error:
        print(error)
        raise
    except json.JSONDecodeError:
        print("Failed to parse SecretString to JSON")
        raise
    for file_name, attribute in AcmeCertificate.files.items():
        file_path = f"{destination_folder}/{file_name}"
        with open(file_path, "w") as fd:
            fd.write(secret_value[attribute])
        set_certs_permissions(
            file_name,
            file_path,
            private_file_name or AcmeCertificate.private_key_file_name,
        )


def pull_from_secretsmanager_secret_per_cert(
    destination_folder: str,
    locations: dict,
    private_file_name: str = None,
    session: Session = None,
) -> None:
    from certbot_aws_store.certificate import AcmeCertificate

    session = get_session(session)
    client = session.client("secretsmanager")
    for file_name, attribute in AcmeCertificate.files.items():
        file_path = f"{destination_folder}/{file_name}"
        try:
            with open(file_path, "w") as fd:
                secret_value = client.get_secret_value(SecretId=locations[attribute])[
                    "SecretString"
                ]
                fd.write(secret_value)
            set_certs_permissions(
                file_name,
                file_path,
                private_file_name or AcmeCertificate.private_key_file_name,
            )
        except client.exceptions.ResourceNotFoundException as error:
            print(error)
            raise


def pull_from_s3(
    destination_folder: str,
    locations: dict,
    private_file_name: str = None,
    session: Session = None,
):
    from certbot_aws_store.certificate import AcmeCertificate

    session = get_session(session)
    for file_name, attribute in AcmeCertificate.files.items():
        file_path = f"{destination_folder}/{file_name}"
        parts = S3Backend.bucket_file_arn_re.match(locations[attribute])
        if not parts:
            raise ValueError(
                locations[attribute],
                "is not a valid S3 ARN",
                S3Backend.bucket_file_arn_re.pattern,
            )
        resource = session.resource("s3").Object(
            parts.group("bucket"), parts.group("key")
        )
        resource.download_file(file_path)
        set_certs_permissions(
            file_name,
            file_path,
            private_file_name or AcmeCertificate.private_key_file_name,
        )


def set_certs_permissions(
    file_name: str, file_path: str, private_file_name: str
) -> None:
    if file_name == private_file_name:
        chmod(
            file_path,
            stat.S_IWRITE | stat.S_IRUSR,
        )
    else:
        chmod(
            file_path,
            stat.S_IWRITE | stat.S_IRUSR | stat.S_IRGRP | stat.S_IROTH,
        )
